#include "jannex/api/console.h"
#include "jannex/jannex.h"

#ifdef CONFIG_HTTP_SERVER_ENABLE_DEBUG_MODE

#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "freertos/semphr.h"
#include "freertos/event_groups.h"

#include "esp_websocket_client.h"
#include "esp_event.h"

#include <stdio.h>
#include <string.h>
#include <esp_log.h>
#include <esp_system.h>

#include "esp_app_desc.h"
#include "esp_mac.h"

#include "mbedtls/base64.h"

#include "esp_partition.h"

#define CALC_BASE64_ENCODED_LENGTH(input_length) ((4 * ((input_length + 2) / 3)) + 1)

char cacheMessage[CONFIG_HTTP_SERVER_DEBUG_CACHE_LENGTH];
char cacheMessageClone[CONFIG_HTTP_SERVER_DEBUG_CACHE_LENGTH];
int cacheMessageIndex = 0;

SemaphoreHandle_t cacheMutex;

const static char *TAG = "ConsoleWS";

esp_websocket_client_handle_t client = NULL;

bool shouldSendCoreDump = false;
int skippedMessages = 0;

static const size_t BLOCK_SIZE = 1024;

void sendCoreDump()
{
    if (shouldSendCoreDump)
    {
        shouldSendCoreDump = false;

        esp_websocket_client_send_text(client, "!DMSG_START!", 12, 10000 / portTICK_PERIOD_MS);

        esp_partition_t *coredump_partition = esp_partition_find_first(
            ESP_PARTITION_TYPE_DATA, ESP_PARTITION_SUBTYPE_DATA_COREDUMP, "coredump");

        if (coredump_partition == NULL)
        {
            ESP_LOGE(TAG, "Core dump partition not found.");
            goto END;
        }

        uint8_t *buffer = (uint8_t *)malloc(BLOCK_SIZE);

        size_t remaining = coredump_partition->size;

        // size_t base64_length = CALC_BASE64_ENCODED_LENGTH(BLOCK_SIZE) + 1;
        // char *base64 = (char *)malloc(base64_length + 1 + 1); // + \n + \0
        // size_t encoded_length = 0;

        while (remaining > 0)
        {
            // wait until wifi is connected
            // xEventGroupWaitBits(s_wifi_event_group, NETWORK_CONNECTED_BIT, false, true, portMAX_DELAY);

            // wait until console is connected
            if (client == NULL || !esp_websocket_client_is_connected(client))
            {
                vTaskDelay(1000 / portTICK_PERIOD_MS);
                continue;
            }

            size_t to_read = (remaining < BLOCK_SIZE) ? remaining : BLOCK_SIZE;
            esp_err_t err = esp_partition_read(coredump_partition, coredump_partition->size - remaining, buffer, to_read);
            if (err != ESP_OK)
            {
                ESP_LOGE(TAG, "Failed to read core dump data.");
                free(buffer);
                // free(base64);

                goto END;
            }

            /*if (base64 != NULL)
            {
                int ret = mbedtls_base64_encode((unsigned char *)base64, base64_length - 1, &encoded_length, (unsigned char *)buffer, to_read);

                if (ret != 0)
                {
                    ESP_LOGE(TAG, "mbedtls_base64_encode failed!");
                    free(buffer);
                    free(base64);

                    goto END;
                }

                // set last char to \n
                base64[encoded_length] = '\n';
                // base64[encoded_length + 1] = '\0';
            }
            else
            {
                ESP_LOGE(TAG, "Memory allocation failed!");
                free(buffer);
                free(base64);

                goto END;
            }*/

            esp_websocket_client_send_bin(client, (const char *)buffer, to_read, 10000 / portTICK_PERIOD_MS);
            remaining -= to_read;

            // vTaskDelay(1);
            printf("free heap: %lu\n", esp_get_free_heap_size());
        }

        esp_websocket_client_send_text(client, "!DMSG_END!", 10, 10000 / portTICK_PERIOD_MS);

        free(buffer);
        // free(base64);
    }

END:
}

void sendOlderMessages()
{
    if (esp_websocket_client_is_connected(client))
    {
        // printf("cacheMessageIndex:%i\n", cacheMessageIndex);

        int newDataLength = 0;

        // mutex lock
        if (xSemaphoreTake(cacheMutex, (TickType_t)10) == pdTRUE)
        {
            if (cacheMessageIndex > 0)
            {
                // clone cacheMessage to cacheMessageClone
                memcpy(cacheMessageClone, cacheMessage, cacheMessageIndex);

                newDataLength = cacheMessageIndex;

                // clear cacheMessage
                cacheMessageIndex = 0;
            }

            // mutex unlock
            xSemaphoreGive(cacheMutex);
        }

        if (newDataLength > 0)
        {

            cacheMessageClone[newDataLength - 1] = '\0';
            esp_websocket_client_send_text(client, cacheMessageClone, newDataLength, 1000 / portTICK_PERIOD_MS);
            newDataLength = 0;

            /*// count \n in cacheMessage
            int count = 0;
            for (int i = 0; i < cacheMessageIndex; i++)
            {
                if (cacheMessage[i] == '\n')
                {
                    count++;
                }
            }

            printf("count:%i\n", count);

            // split cacheMessage into multiple messages
            char *pch = strtok(cacheMessage, "\n");
            while (pch != NULL)
            {
                if (strlen(pch) > 0)
                {
                    printf("pf-'%s'\n", pch);

                    esp_websocket_client_send_text(client, pch, strlen(pch), 1000 / portTICK_PERIOD_MS);
                }
                pch = strtok(NULL, "\n");
            }
            cacheMessageIndex = 0;*/
        }
        else
        {
        }
    }
    else
    {
    }
}

int jannex_LOG(const char *format, va_list args)
{
    int status = 0;

    //  sprintf to jannex_LOG func and dynamically allocate memory
    char *esp_message;
    status = vasprintf(&esp_message, format, args);

    size_t encoded_length = 0;
    size_t input_length = strlen(esp_message);
    size_t message_length = CALC_BASE64_ENCODED_LENGTH(input_length) + 1;

    char *message = (char *)malloc(message_length + 1 + 1); // + \n + \0

    if (message != NULL)
    {
        int ret = mbedtls_base64_encode((unsigned char *)message, message_length - 1, &encoded_length, (unsigned char *)esp_message, input_length);

        if (ret != 0)
        {
            printf("Base64 encoding failed!");
            skippedMessages++;

            goto END;
        }

        // set last char to \n
        message[encoded_length] = '\n';
        message[encoded_length + 1] = '\0';
    }
    else
    {
        printf("Memory allocation failed!");
        skippedMessages++;

        goto END;
    }

    message_length = encoded_length + 1;

    // check if socket is online
    /*if (client != NULL && esp_websocket_client_is_connected(client))
    {
        // print last char of message in hex
        // printf("last char: %02x\n", message[encoded_length - 1]);

        sendOlderMessages();

        printf("message_length:%i\n", message_length);
        esp_websocket_client_send_text(client, message, message_length, 1000 / portTICK_PERIOD_MS);
    }
    else*/
    {
        // Take the mutex before writing to the variable
        if (xSemaphoreTake(cacheMutex, portMAX_DELAY) == pdTRUE)
        {

            if (cacheMessageIndex + message_length + 1 > CONFIG_HTTP_SERVER_DEBUG_CACHE_LENGTH)
            {
                // Cache is full
                // replace at the end of the cache

                strcpy(cacheMessage + CONFIG_HTTP_SERVER_DEBUG_CACHE_LENGTH - 7, "!CCO!\n");
                cacheMessageIndex = CONFIG_HTTP_SERVER_DEBUG_CACHE_LENGTH;

                xSemaphoreGive(cacheMutex);
                goto END;
            }

            // cache message thread safe
            // xSemaphoreTake(cacheMessageSemaphore, portMAX_DELAY);

            // cache message
            if (cacheMessageIndex == 0)
            {
                strcpy(cacheMessage + cacheMessageIndex, message);
                cacheMessageIndex += message_length;
            }
            else
            {
                // cacheMessage[cacheMessageIndex] = '\n';
                strcpy(cacheMessage + cacheMessageIndex, message);
                cacheMessageIndex += message_length;
            }

            // Release the mutex
            xSemaphoreGive(cacheMutex);
        }
        else
        {
            // Unable to take the mutex
            // Handle the error
        }
    }

END:

    free(esp_message);
    free(message);

    if (status >= 0)
    {

        return vprintf(format, args); // print to console
    }
    return status;
}

static void log_error_if_nonzero(const char *message, int error_code)
{
    if (error_code != 0)
    {
        ESP_LOGE(TAG, "Last error %s: 0x%x", message, error_code);
    }
}

static void websocket_event_handler(void *handler_args, esp_event_base_t base, int32_t event_id, void *event_data)
{
    esp_websocket_event_data_t *data = (esp_websocket_event_data_t *)event_data;
    switch (event_id)
    {
    case WEBSOCKET_EVENT_CONNECTED:
        ESP_LOGI(TAG, "WEBSOCKET_EVENT_CONNECTED");
        sendOlderMessages();
        printf("sending core dump\n");
        sendCoreDump();
        printf("sending core dump done\n");
        // xTaskCreate(sendCoreDumpTask, "sendCoreDumpTask", 4096, NULL, 5, NULL);

        ESP_LOGI(TAG, "Websocket connected! %s", CONFIG_HTTP_SERVER_DEBUG_URL);
        break;
    case WEBSOCKET_EVENT_DISCONNECTED:
        ESP_LOGE(TAG, "Websocket disconnected!");
        ESP_LOGI(TAG, "WEBSOCKET_EVENT_DISCONNECTED");
        log_error_if_nonzero("HTTP status code", data->error_handle.esp_ws_handshake_status_code);
        if (data->error_handle.error_type == WEBSOCKET_ERROR_TYPE_TCP_TRANSPORT)
        {
            log_error_if_nonzero("reported from esp-tls", data->error_handle.esp_tls_last_esp_err);
            log_error_if_nonzero("reported from tls stack", data->error_handle.esp_tls_stack_err);
            log_error_if_nonzero("captured as transport's socket errno", data->error_handle.esp_transport_sock_errno);
        }
        break;
    case WEBSOCKET_EVENT_DATA:
        if (data->payload_len == 0)
        {
            break;
        }

        ESP_LOGI(TAG, "Received opcode=%d", data->op_code);
        if (data->op_code == 0x08 && data->data_len == 2)
        {
            ESP_LOGW(TAG, "Received closed message with code=%d", 256 * data->data_ptr[0] + data->data_ptr[1]);
        }
        else
        {
            ESP_LOGI(TAG, "Received=%.*s", data->data_len, (char *)data->data_ptr);

            // check if message is coredumpOK
            if (data->data_len == 10 && strncmp((char *)data->data_ptr, "coredumpOK", 10) == 0)
            {
                // delete coredump partition
                esp_partition_t *coredump_partition = esp_partition_find_first(
                    ESP_PARTITION_TYPE_DATA, ESP_PARTITION_SUBTYPE_DATA_COREDUMP, "coredump");

                if (coredump_partition == NULL)
                {
                    ESP_LOGE(TAG, "Core dump partition not found.");
                    break;
                }

                esp_err_t err = esp_partition_erase_range(coredump_partition, 0, coredump_partition->size);
                if (err != ESP_OK)
                {
                    ESP_LOGE(TAG, "Failed to erase core dump partition.");
                    break;
                }

                // set coredump was downloaded
                // setCoredumpWasDownloaded(true);

                ESP_LOGI(TAG, "Core dump partition erased.");
            }
        }

        // If received data contains json structure it succeed to parse
        // ESP_LOGW(TAG, "Total payload length=%d, data_len=%d, current payload offset=%d\r\n", data->payload_len, data->data_len, data->payload_offset);*/

        break;
    case WEBSOCKET_EVENT_ERROR:
        ESP_LOGI(TAG, "WEBSOCKET_EVENT_ERROR");
        log_error_if_nonzero("HTTP status code", data->error_handle.esp_ws_handshake_status_code);
        if (data->error_handle.error_type == WEBSOCKET_ERROR_TYPE_TCP_TRANSPORT)
        {
            log_error_if_nonzero("reported from esp-tls", data->error_handle.esp_tls_last_esp_err);
            log_error_if_nonzero("reported from tls stack", data->error_handle.esp_tls_stack_err);
            log_error_if_nonzero("captured as transport's socket errno", data->error_handle.esp_transport_sock_errno);
        }
        break;
    }
}

// function like printf and redirect the output to websocket

static void console_task(void *pvParameters)
{

    // get mac address
    uint8_t mac[6];
    esp_efuse_mac_get_default(mac);

    // get current version
    const esp_app_desc_t *app_desc = esp_app_get_description();

    char branch[32];
    getBranch(branch);

    char url[256];

    sprintf(url, "%s?mac=%02X:%02X:%02X:%02X:%02X:%02X&device=%s&branch=%s&ver=%s", CONFIG_HTTP_SERVER_DEBUG_URL, mac[0], mac[1], mac[2], mac[3], mac[4], mac[5], app_desc->project_name, branch, app_desc->version);

    esp_websocket_client_config_t websocket_cfg = {
        .uri = url,
        .port = CONFIG_HTTP_SERVER_DEBUG_PORT,

        .task_stack = 1024 * 6,
        .reconnect_timeout_ms = 5000,
        .network_timeout_ms = 500,
        .ping_interval_sec = 10000,
    };

    ESP_LOGI(TAG, "Connecting to %s... on port %d", websocket_cfg.uri, websocket_cfg.port);

    client = esp_websocket_client_init(&websocket_cfg);
    esp_websocket_register_events(client, WEBSOCKET_EVENT_ANY, websocket_event_handler, (void *)client);

    esp_websocket_client_start(client);

    // just test here some data
    /*char data[32];
    int i = 0;
    while (true)
    {
        int len = sprintf(data, "hello %04d", i++);
        ESP_LOGI(TAG, "Sending %s", data);

        vTaskDelay(1000 / portTICK_PERIOD_MS);
    }

    esp_websocket_client_destroy(client);*/

    while (true)
    {
        sendOlderMessages();

        vTaskDelay(500 / portTICK_PERIOD_MS);
    }
}

void initConsole()
{
    cacheMutex = xSemaphoreCreateMutex();
    if (cacheMutex == NULL)
    {
        // Mutex creation failed
        // Handle the error
        printf("initConsole Mutex creation failed!!!!!!!!!!!");
    }

    esp_log_set_vprintf(&jannex_LOG);

    if (cacheMutex == NULL)
    {
        // Mutex creation failed
        // Handle the error
        printf("initConsole Mutex creation failed!!!!!!!!!!!");
        ESP_LOGE(TAG, "initConsole Mutex creation failed!!!!!!!!!!!");
    }
}

void connectConsole()
{
    // wait unitl wifi is connected
    // xEventGroupWaitBits(s_wifi_event_group, NETWORK_CONNECTED_BIT, false, true, portMAX_DELAY);
    vTaskDelay(3000 / portTICK_PERIOD_MS);

    xTaskCreate(console_task, "console_task", 4096, NULL, 5, NULL);
}

void sendCoredumpToConsole()
{
    shouldSendCoreDump = true;
}

#endif // CONFIG_HTTP_SERVER_ENABLE_DEBUG_MODE
