#include "jannex/jannex.h"

#include <stdio.h>
#include <string.h>
#include "esp_log.h"
#include "esp_partition.h"
#include "esp_http_server.h"

#include <esp_http_server.h>

static const char *TAG = "coredumpServer";

static const size_t BLOCK_SIZE = 1024; // Adjust the block size as needed

RTC_NOINIT_ATTR uint16_t panicRetryCounter;
RTC_NOINIT_ATTR uint16_t softwareResetCounter;

bool checkIfCoredumpWasDownloaded()
{
    return false;
}

bool setCoredumpWasDownloaded(bool value)
{
    return true;
}

esp_err_t coredump_get_handler(httpd_req_t *req)
{
    esp_partition_t *coredump_partition = esp_partition_find_first(
        ESP_PARTITION_TYPE_DATA, ESP_PARTITION_SUBTYPE_DATA_COREDUMP, "coredump");

    if (coredump_partition == NULL)
    {
        ESP_LOGE(TAG, "Core dump partition not found.");
        httpd_resp_send_err(req, HTTPD_500_INTERNAL_SERVER_ERROR, "Core dump not available");
        return ESP_FAIL;
    }

    // Set the HTTP response headers for downloading a .bin file
    httpd_resp_set_type(req, "application/octet-stream");
    httpd_resp_set_hdr(req, "Content-Disposition", "attachment; filename=coredump.bin");

    uint8_t *buffer = (uint8_t *)malloc(BLOCK_SIZE);
    if (buffer == NULL)
    {
        ESP_LOGE(TAG, "Memory allocation failed.");
        httpd_resp_send_err(req, HTTPD_500_INTERNAL_SERVER_ERROR, "Memory allocation failed");
        return ESP_FAIL;
    }

    size_t remaining = coredump_partition->size;
    while (remaining > 0)
    {
        size_t to_read = (remaining < BLOCK_SIZE) ? remaining : BLOCK_SIZE;
        esp_err_t err = esp_partition_read(coredump_partition, coredump_partition->size - remaining, buffer, to_read);
        if (err != ESP_OK)
        {
            ESP_LOGE(TAG, "Failed to read core dump data.");
            free(buffer);
            httpd_resp_send_err(req, HTTPD_500_INTERNAL_SERVER_ERROR, "Failed to read core dump data");
            return ESP_FAIL;
        }

        // Send a block of data as the HTTP response
        httpd_resp_send_chunk(req, (const char *)buffer, to_read);

        remaining -= to_read;
    }

    free(buffer);
    httpd_resp_send_chunk(req, NULL, 0); // Send an empty chunk to finish the response
    return ESP_OK;
}

httpd_uri_t coreDumpRequest = {
    .uri = "/getCoreDumpData",
    .method = HTTP_GET,
    .handler = coredump_get_handler,
    .user_ctx = NULL};
